# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "17/11/2020"

import unittest
import numpy

from darfix.core import mapping


class TestMapping(unittest.TestCase):

    """Tests for `mapping.py`"""

    def setUp(self):
        self.data = numpy.random.random(size=(3, 10, 10))

    def test_generator1(self):
        """Tests the correct creation of a generator without moments"""

        g = mapping.generator(self.data)

        img, moment = next(g)
        self.assertEqual(moment, None)
        numpy.testing.assert_array_equal(img, self.data[:, 0, 0])

    def test_generator2(self):
        """Tests the correct creation of a generator with moments"""

        moments = numpy.ones((3, 10, 10))
        g = mapping.generator(self.data, moments)

        img, moment = next(g)
        numpy.testing.assert_array_equal(moment, moments[:, 0, 0])
        numpy.testing.assert_array_equal(img, self.data[:, 0, 0])

    def test_fit_rocking_curve(self):
        """Tests the correct fit of a rocking curve"""

        samples = numpy.random.normal(size=10000) + numpy.random.random(10000)

        y, bins = numpy.histogram(samples, bins=100)

        y_pred, pars = mapping.fit_rocking_curve([y, None])
        rss = numpy.sum((y - y_pred) ** 2)
        tss = numpy.sum((y - y.mean()) ** 2)
        r2 = 1 - rss / tss

        self.assertGreater(r2, 0.9)
        self.assertEqual(len(pars), 4)

    def test_fit_data(self):
        """Tests the new data has same shape as initial data"""

        new_data, maps = mapping.fit_data(self.data)

        self.assertEqual(new_data.shape, self.data.shape)
        self.assertEqual(len(maps), 4)
        self.assertEqual(maps[0].shape, self.data[0].shape)

    def test_moments(self):
        """Tests the correct moments calculation"""

        values = [0.1, 0.2, 0.3]

        com, std, skews, kurt = mapping.compute_moments(values, self.data)

        self.assertEqual(com.shape, self.data.shape[1:])
        self.assertEqual(std.shape, self.data.shape[1:])
        self.assertEqual(skews.shape, self.data.shape[1:])
        self.assertEqual(kurt.shape, self.data.shape[1:])

    def test_rsm(self):
        """Tests RSM"""

        H, W = self.data.shape[1:]
        d = 0.1
        ffz = 10
        mainx = 5

        pix_arr = mapping.compute_rsm(H, W, d, ffz, mainx)

        self.assertEqual(pix_arr[0].shape, (H, W))
        self.assertEqual(pix_arr[1].shape, (H, W))

    def test_magnification(self):
        """Tests magnification"""

        H, W = self.data.shape[1:]
        d = 0.1
        obx = 10
        obpitch = 25.1
        mainx = 5

        pix_arr = mapping.compute_magnification(H, W, d, obx, obpitch, mainx)

        self.assertEqual(pix_arr[0].shape, (H, W))
        self.assertEqual(pix_arr[1].shape, (H, W))

    def test_magnification_uncentered(self):
        """Tests magnification uncentered"""

        H, W = self.data.shape[1:]
        d = 0.1
        obx = 10
        obpitch = 25.1
        mainx = 50  # Has to be big enough

        pix_arr = mapping.compute_magnification(
            H, W, d, obx, obpitch, mainx, center=False
        )

        self.assertEqual(pix_arr[0][0][0], 0)
        self.assertEqual(pix_arr[1][H - 1][0], 0)

    def test_peak_position(self):
        """Tests peak position map"""

        image = mapping.compute_peak_position(self.data)

        self.assertEqual(image[0, 1], numpy.argmax(self.data[:, 0, 1]))

    def test_peak_position_values(self):
        """Tests peak position map with values"""

        values = numpy.repeat([numpy.linspace(0.1, 1, 10)], 3).flatten()
        image = mapping.compute_peak_position(self.data, values)

        self.assertEqual(image[0, 0], values[numpy.argmax(self.data[:, 0, 0])])



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "03/12/2020"

import glob
import os

from silx.io import fabioh5
from silx.io.url import DataUrl
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.StackView import StackViewMainWindow

import darfix
from darfix.core.dataset import Data
from darfix.gui.datasetSelectionWidget import DirSelectionWidget


class ShowStackWidget(qt.QMainWindow):
    """
    Widget to show a stack of data
    """

    sigComputed = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        layout = qt.QVBoxLayout()
        self._filenameData = DirSelectionWidget(parent=self)
        self._filenameData.dirChanged.connect(self.updateStack)
        layout.addWidget(self._filenameData)
        self._sv = StackViewMainWindow()
        self._sv.setColormap(
            Colormap(name=darfix.config.DEFAULT_COLORMAP_NAME, normalization="linear")
        )
        self._sv.hide()
        layout.addWidget(self._sv)
        widget = qt.QWidget(parent=self)
        widget.setLayout(layout)

        self.setCentralWidget(widget)

    def updateStack(self):
        """
        Update stack with new first filename
        """
        _dir = self._filenameData.getDir()
        data_urls = []
        metadata = []
        for filename in sorted(glob.glob(os.path.join(_dir, "*"))):
            if os.path.isfile(filename):
                data_urls.append(DataUrl(file_path=filename, scheme="fabio"))
                metadata.append(fabioh5.EdfFabioReader(file_name=filename))

        data = Data(urls=data_urls, metadata=metadata, in_memory=False)

        self._sv.setStack(data)
        self._sv.show()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "22/04/2020"

import numpy
import unittest

from darfix.decomposition.base import Base


class TestBase(unittest.TestCase):
    """Tests for `base.py`."""

    def setUp(self):
        self.images = numpy.random.random((100, 1000))

    def test_data(self):
        base = Base(self.images)
        numpy.testing.assert_equal(self.images, base.data)

    def test_indices(self):
        base = Base(self.images)
        numpy.testing.assert_equal(base.indices, numpy.arange(len(self.images)))

        base = Base(self.images, indices=numpy.arange(20))
        numpy.testing.assert_equal(base.indices, numpy.arange(20))

    def test_num_components(self):
        base = Base(self.images)
        self.assertEqual(base.num_components, 100)

        base = Base(self.images, num_components=10)
        self.assertEqual(base.num_components, 10)

    def test_W(self):
        base = Base(self.images, num_components=10)
        base.fit_transform()

        self.assertEqual(base.W.shape, (100, 10))

    def test_H(self):
        base = Base(self.images, num_components=10)
        base.fit_transform()

        self.assertEqual(base.H.shape, (10, 1000))

    def test_fit_transform(self):
        base = Base(self.images)
        base.fit_transform(compute_w=False)
        self.assertFalse(hasattr(base, "W"))
        self.assertTrue(hasattr(base, "H"))

        base = Base(self.images)
        base.fit_transform(compute_h=False)
        self.assertFalse(hasattr(base, "H"))
        self.assertTrue(hasattr(base, "W"))

    def test_frobenius_norm(self):
        base = Base(self.images)
        self.assertEqual(base.frobenius_norm(), None)

        base.fit_transform()
        self.assertNotEqual(base.frobenius_norm(), None)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "16/06/2021"


import unittest

import numpy
import tempfile
import shutil

try:
    import scipy
except ImportError:
    scipy = None

from darfix.core import imageRegistration
from darfix.test import utils
from darfix.core.dimension import POSITIONER_METADATA


class TestImageRegistration(unittest.TestCase):

    """Tests for `imageRegistration.py`."""

    @classmethod
    def setUpClass(cls):
        cls.rstate = numpy.random.RandomState(seed=1000)
        cls.data = numpy.array(
            [
                [
                    [1, 2, 3, 4, 5],
                    [2, 2, 3, 4, 5],
                    [3, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 3],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [8, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
            ]
        )

    def test_find_shift(self):
        """Tests the shift found"""
        shift = (-1.4, 1.32)
        for img in self.data:
            # The shift corresponds to the pixel offset relative to the reference image
            offset_image = imageRegistration._opencv_fft_shift(img, shift[1], shift[0])
            computed_shift = imageRegistration.find_shift(img, offset_image, 100)
            numpy.testing.assert_allclose(shift, -computed_shift, rtol=1e-04)

    def test_apply_shift(self):
        """Tests the correct apply of the shift"""
        shift = (0, 0)
        for img in self.data:
            shifted_image = imageRegistration.apply_shift(img, shift)
            numpy.testing.assert_allclose(img, shifted_image, rtol=1e-04)
        for img in self.data:
            shifted_image = imageRegistration.apply_shift(
                img, shift, shift_approach="linear"
            )
            numpy.testing.assert_allclose(img, shifted_image, rtol=1e-04)

    def test_improve_shift(self):
        """Tests the shift improvement"""
        h = imageRegistration.improve_linear_shift(
            self.data, [1, 1], 0.1, 0.1, 1, shift_approach="fft"
        )
        numpy.testing.assert_allclose(h, 0.0)
        h = imageRegistration.improve_linear_shift(
            self.data, [1, 1], 0.1, 0.1, 1, shift_approach="linear"
        )
        numpy.testing.assert_allclose(h, 0.0)

    @unittest.skipUnless(scipy, "scipy is missing")
    def test_shift_detection10(self):
        """Tests the shift detection with tolerance of 3 decimals"""
        first_frame = numpy.zeros((100, 100))
        # Simulating a series of frame with information in the middle.
        first_frame[25:75, 25:75] = self.rstate.randint(50, 300, size=(50, 50))
        data = [first_frame]
        shift = [1.0, 0]
        for i in range(9):
            data += [
                numpy.fft.ifftn(
                    scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
                ).real
            ]
        data = numpy.asanyarray(data, dtype=numpy.int16)
        optimal_shift = imageRegistration.shift_detection(
            data, 100, shift_approach="fft"
        )

        shift = [
            [0, -1, -2, -3, -4, -5, -6, -7, -8, -9],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]

        numpy.testing.assert_allclose(shift, numpy.round(optimal_shift))

    @unittest.skipUnless(scipy, "scipy is missing")
    def test_shift_detection01(self):
        """Tests the shift detection with tolerance of 5 decimals"""
        # Create a frame and repeat it shifting it every time
        first_frame = numpy.zeros((100, 100))
        first_frame[25:75, 25:75] = self.rstate.randint(50, 300, size=(50, 50))
        data = [first_frame]
        shift = [0, 1]
        for i in range(9):
            data += [
                numpy.fft.ifftn(
                    scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
                ).real
            ]
        data = numpy.asanyarray(data, dtype=numpy.int16)
        optimal_shift = imageRegistration.shift_detection(
            data, 100, shift_approach="fft"
        )

        shift = [
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, -1, -2, -3, -4, -5, -6, -7, -8, -9],
        ]

        numpy.testing.assert_allclose(shift, numpy.round(optimal_shift))

    @unittest.skipUnless(scipy, "scipy is missing")
    def test_shift_detection11(self):
        """Tests the shift detection with tolerance of 2 decimals"""
        # Create a frame and repeat it shifting it every time
        first_frame = numpy.zeros((100, 100))
        first_frame[25:75, 25:75] = self.rstate.randint(50, 300, size=(50, 50))
        data = [first_frame]
        shift = [1, 1]
        for i in range(9):
            data += [
                numpy.fft.ifftn(
                    scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
                ).real
            ]
        data = numpy.asanyarray(data, dtype=numpy.int16)

        optimal_shift = imageRegistration.shift_detection(data, 100)

        shift = [
            [0, -1, -2, -3, -4, -5, -6, -7, -8, -9],
            [0, -1, -2, -3, -4, -5, -6, -7, -8, -9],
        ]

        numpy.testing.assert_allclose(shift, numpy.round(optimal_shift))

    @unittest.skipUnless(scipy, "scipy is missing")
    def test_shift_detection_float(self):
        """Tests the shift detection using shifted float with tolerance of 2 decimals"""
        first_frame = numpy.zeros((100, 100))
        # Simulating a series of frame with information in the middle.
        first_frame[25:75, 25:75] = self.rstate.randint(50, 300, size=(50, 50))
        data = [first_frame]
        shift = [0.5, 0.2]
        for i in range(9):
            data += [
                numpy.fft.ifftn(
                    scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
                ).real
            ]
        data = numpy.asanyarray(data, dtype=numpy.int16)
        optimal_shift = imageRegistration.shift_detection(data, 100)
        shift = [
            [0, -0.5, -1, -1.5, -2, -2.5, -3, -3.5, -4, -4.5],
            [0, -0.2, -0.4, -0.6, -0.8, -1, -1.2, -1.4, -1.6, -1.8],
        ]

        numpy.testing.assert_allclose(shift, numpy.round(optimal_shift, 1))

    def test_shift_correction00(self):
        """Tests the shift correction of a [0,0] shift."""

        data = imageRegistration.shift_correction(
            self.data, numpy.outer([0, 0], numpy.arange(3))
        )
        numpy.testing.assert_allclose(data, self.data, rtol=1e-03)

    def test_shift_correction01(self):
        """Tests the shift correction of a [0,1] shift."""

        expected = numpy.array(
            [
                [
                    [1, 2, 3, 4, 5],
                    [2, 2, 3, 4, 5],
                    [3, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 3],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [8, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
            ]
        )

        data = imageRegistration.shift_correction(
            self.data, numpy.outer([1, 0], numpy.arange(3))
        )
        numpy.testing.assert_allclose(data, expected, rtol=1e-03)

    def test_shift_correction10(self):
        """Tests the shift correction of a [1,0] shift."""

        expected = numpy.array(
            [
                [
                    [1, 2, 3, 4, 5],
                    [2, 2, 3, 4, 5],
                    [3, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [5, 1, 2, 3, 4],
                    [5, 1, 2, 3, 4],
                    [3, 1, 2, 3, 4],
                    [5, 1, 2, 3, 4],
                    [5, 1, 2, 3, 4],
                ],
                [
                    [4, 5, 1, 2, 3],
                    [4, 5, 1, 2, 3],
                    [4, 5, 1, 2, 3],
                    [4, 5, 8, 2, 3],
                    [4, 5, 1, 2, 3],
                ],
            ]
        )

        data = imageRegistration.shift_correction(
            self.data, numpy.outer([0, 1], numpy.arange(3))
        )
        numpy.testing.assert_allclose(data, expected, rtol=1e-05)

    def test_shift_correction11(self):
        """Tests the shift correction of a [1,1] shift."""

        expected = numpy.array(
            [
                [
                    [1, 2, 3, 4, 5],
                    [2, 2, 3, 4, 5],
                    [3, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [5, 1, 2, 3, 4],
                    [5, 1, 2, 3, 4],
                    [5, 1, 2, 3, 4],
                    [3, 1, 2, 3, 4],
                    [5, 1, 2, 3, 4],
                ],
                [
                    [4, 5, 8, 2, 3],
                    [4, 5, 1, 2, 3],
                    [4, 5, 1, 2, 3],
                    [4, 5, 1, 2, 3],
                    [4, 5, 1, 2, 3],
                ],
            ]
        )
        data = imageRegistration.shift_correction(
            self.data, numpy.outer([1, 1], numpy.arange(3))
        )
        numpy.testing.assert_allclose(data, expected, rtol=1e-05)

    def test_shift_correction_float(self):
        """Tests the shift correction of a [0.1, 0.25] shift between images."""

        expected = [
            [
                [1, 2, 3, 4, 5],
                [2, 2, 3, 4, 5],
                [3, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
            [
                [1.2595824, 1.7349374, 3.0299857, 3.756984, 4.9321423],
                [1.3387496, 1.6893137, 3.0737815, 3.690435, 5.6077204],
                [1.0840675, 1.836086, 2.9328897, 3.904524, 3.4343739],
                [1.220753, 1.7573147, 3.008505, 3.7896245, 4.600788],
                [1.3292272, 1.6948014, 3.0685136, 3.6984396, 5.5264597],
            ],
            [
                [0.03080546, 1.01156497, 3.30827889, 3.27796478, 5.64089073],
                [2.96707199, 1.77546528, 2.90155831, 3.6526126, 5.10329182],
                [0.03080546, 1.01156497, 3.30827889, 3.27796478, 5.64089073],
                [5.90333852, 2.53936558, 2.49483774, 4.02726042, 4.56569291],
                [5.90333852, 2.53936558, 2.49483774, 4.02726042, 4.56569291],
            ],
        ]

        data = imageRegistration.shift_correction(
            self.data, numpy.outer([0.25, 0.1], numpy.arange(3))
        )
        numpy.testing.assert_allclose(data, expected, rtol=1e-05)


class TestReshapedShift(unittest.TestCase):
    def setUp(self):
        """ "
        Creating random dataset with specific headers.
        """
        self._dir = tempfile.mkdtemp()
        counter_mne = "a b c d e f g h"
        motor_mne = "x y z k h m n"
        # Create headers
        header = []
        # Dimensions for reshaping
        self.rstate = numpy.random.RandomState(seed=1000)
        self.first_dim = self.rstate.rand(5)
        self.second_dim = self.rstate.rand(2)
        motors = self.rstate.rand(7)
        for i in numpy.arange(10):
            header.append({})
            header[i]["HeaderID"] = i
            header[i]["counter_mne"] = counter_mne
            header[i]["motor_mne"] = motor_mne
            header[i]["counter_pos"] = ""
            header[i]["motor_pos"] = ""
            for c in counter_mne:
                header[i]["counter_pos"] += str(self.rstate.rand(1)[0]) + " "
            for j, m in enumerate(motor_mne.split()):
                if m == "z":
                    header[i]["motor_pos"] += str(self.first_dim[i % 5]) + " "
                elif m == "m":
                    header[i]["motor_pos"] += str(self.second_dim[int(i > 4)]) + " "
                else:
                    header[i]["motor_pos"] += str(motors[j]) + " "

        self.header = header
        self.first_frame = numpy.zeros((100, 100))
        self.first_frame[30:40, 30:40] = self.rstate.randint(50, 100, size=(10, 10))

    def test_shift_detection0(self):
        """Tests the shift detection using only an axis (dimension).
        The shift is only applied to the dimension."""
        data = [self.first_frame]
        shift = [0.5, 0.2]
        for i in range(1, 10):
            if i < 5:
                data += [
                    numpy.fft.ifftn(
                        scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
                    ).real
                ]
            else:
                data += [data[-1]]
        data = numpy.asanyarray(data, dtype=numpy.int16)
        self.dataset = utils.createDataset(
            data=data, header=self.header, _dir=self._dir
        )

        self.dataset.find_dimensions(POSITIONER_METADATA)
        dataset = self.dataset.reshape_data()

        # Detects shift using only images where value 1 of dimension 1 is fixed
        optimal_shift = dataset.find_shift(dimension=[1, 0])

        shift = [[0, -0.5, -1, -1.5, -2], [0, -0.2, -0.4, -0.6, -0.8]]

        numpy.testing.assert_allclose(shift, numpy.round(optimal_shift, 1))

    def test_shift_detection1(self):
        """Tests the shift detection using only an axis (dimension).
        The shift is applied to all the dataset."""
        data = [self.first_frame]
        shift = [0.5, 0.2]
        for i in range(1, 10):
            data += [
                numpy.fft.ifftn(
                    scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
                ).real
            ]
        data = numpy.asanyarray(data, dtype=numpy.int16)
        self.dataset = utils.createDataset(
            data=data, header=self.header, _dir=self._dir
        )

        self.dataset.find_dimensions(POSITIONER_METADATA)
        dataset = self.dataset.reshape_data()

        # Detects shift using only images where value 1 of dimension 1 is fixed
        optimal_shift = dataset.find_shift(dimension=[0, 0])

        shift = [[0, -2.5], [0, -1]]

        numpy.testing.assert_allclose(shift, numpy.round(optimal_shift, 1))

    def test_shift_correction0(self):
        """Tests the shift correction using only an axis (dimension).
        The shift is only applied to the dimension."""
        data = [self.first_frame]
        shift = [0.5, 0.2]
        for i in range(1, 10):
            if i < 5:
                data += [
                    numpy.fft.ifftn(
                        scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
                    ).real
                ]
            else:
                data += [data[-1]]

        data = numpy.asanyarray(data, dtype=numpy.int16)
        self.dataset = utils.createDataset(
            data=data, header=self.header, _dir=self._dir
        )

        self.dataset.find_dimensions(POSITIONER_METADATA)
        dataset = self.dataset.reshape_data()

        dataset = dataset.find_and_apply_shift(dimension=[1, 0])

        for frame in dataset.data.take(0, 0):
            # Check if the difference between the shifted frames and the sample frame is small enough
            self.assertTrue((abs(data[0] - numpy.round(frame, 1)) < 6).all())

    def tearDown(self):
        shutil.rmtree(self._dir)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "06/12/2019"


import unittest

import numpy

from darfix.core import roi


class TestROI(unittest.TestCase):

    """Tests for `roi.py`."""

    @classmethod
    def setUpClass(cls):

        cls.data = numpy.array(
            [
                [
                    [1, 2, 3, 4, 5],
                    [2, 2, 3, 4, 5],
                    [3, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 3],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [8, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
            ]
        )

        cls.dark = [
            [
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
            [
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
            [
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
                [1, 2, 3, 4, 5],
            ],
        ]

    def test_ROI_0(self):
        """Tests the roi of an image."""

        expected = numpy.array([[2, 3, 4], [2, 3, 4], [2, 3, 4]])

        data = roi.apply_2D_ROI(
            self.data[0], size=[3, 3], center=numpy.array(self.data[0].shape) / 2
        )
        numpy.testing.assert_equal(data, expected)

    def test_ROI_1(self):
        """Tests the roi of an image."""

        expected = numpy.array([[1, 2], [2, 2]])

        data = roi.apply_2D_ROI(self.data[0], size=[3, 3], center=[0, 0])
        numpy.testing.assert_equal(data, expected)

    def test_ROI_2(self):
        """Tests the roi of an image."""

        expected = numpy.array([[2, 3, 4, 5], [2, 3, 4, 5], [2, 3, 4, 5], [2, 3, 4, 5]])

        data = roi.apply_2D_ROI(self.data[0], size=[4, 4], center=[2, 3])
        numpy.testing.assert_equal(data, expected)

    def test_ROI_3D_0(self):
        """Tests the roi of an image."""

        expected = numpy.array([[[1, 2], [2, 2]], [[1, 2], [1, 2]], [[1, 2], [1, 2]]])

        data = roi.apply_3D_ROI(self.data, size=[2, 2], origin=[0, 0])
        numpy.testing.assert_equal(data, expected)

    def test_ROI_3D_1(self):
        """Tests the roi of an image."""

        expected = numpy.array([[[1]], [[1]], [[1]]])

        data = roi.apply_3D_ROI(self.data, size=[2, 2], center=[0, 0])
        numpy.testing.assert_equal(data, expected)

    def test_ROI_3D_2(self):
        """Tests the roi of an image."""

        expected = numpy.array(
            [
                [[2, 3, 4, 5], [2, 3, 4, 5], [2, 3, 4, 5], [2, 3, 4, 5]],
                [[2, 3, 4, 5], [2, 3, 4, 5], [2, 3, 4, 3], [2, 3, 4, 5]],
                [[2, 3, 4, 5], [2, 3, 4, 5], [2, 3, 4, 5], [2, 3, 4, 5]],
            ]
        )

        data = roi.apply_3D_ROI(self.data, size=[4, 4], center=[2, 3])
        numpy.testing.assert_equal(data, expected)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "05/07/2021"

# import os
import numpy

from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.StackView import StackViewMainWindow

import darfix
from darfix.core.dataset import Operation

from .operationThread import OperationThread
from .utils import ChooseDimensionDock


class ShiftCorrectionDialog(qt.QDialog):
    """
    Dialog with `ShiftCorrectionWidget` as main window and standard buttons.
    """

    okSignal = qt.Signal()

    def __init__(self, parent=None):
        qt.QDialog.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self._buttons.setEnabled(False)
        resetB = self._buttons.addButton(self._buttons.Reset)
        self.mainWindow = ShiftCorrectionWidget(parent=self)
        self.mainWindow.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(self.mainWindow)
        self.layout().addWidget(self._buttons)

        self._buttons.accepted.connect(self.okSignal.emit)
        resetB.clicked.connect(self.mainWindow.resetStack)
        self.mainWindow.computingSignal.connect(self._toggleButton)

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):

        if dataset is not None:
            self._buttons.setEnabled(True)
            self.mainWindow.setDataset(parent, dataset, indices, bg_indices, bg_dataset)

    def _toggleButton(self, state):
        self._buttons.button(qt.QDialogButtonBox.Ok).setEnabled(not state)


class ShiftCorrectionWidget(qt.QMainWindow):
    """
    A widget to apply shift correction to a stack of images
    """

    computingSignal = qt.Signal(bool)

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        self.setWindowFlags(qt.Qt.Widget)
        self._shift = numpy.array([0, 0])
        self._filtered_shift = None
        self._dimension = None
        self._update_dataset = None
        self.indices = None
        self.bg_indices = None
        self.bg_dataset = None

        self._inputDock = _InputDock()
        self._inputDock.widget.correctionB.setEnabled(False)

        self._sv = StackViewMainWindow()
        self._sv.setColormap(
            Colormap(name=darfix.config.DEFAULT_COLORMAP_NAME, normalization="linear")
        )
        self.setCentralWidget(self._sv)
        self._chooseDimensionDock = ChooseDimensionDock(self)
        spacer1 = qt.QWidget(parent=self)
        spacer1.setLayout(qt.QVBoxLayout())
        spacer1.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        spacer2 = qt.QWidget(parent=self)
        spacer2.setLayout(qt.QVBoxLayout())
        spacer2.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self._chooseDimensionDock.widget.layout().addWidget(spacer1)
        self._inputDock.widget.layout().addWidget(spacer2)
        self._chooseDimensionDock.hide()
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._chooseDimensionDock)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._inputDock)

        self._inputDock.widget.correctionB.clicked.connect(self.correct)
        self._inputDock.widget.abortB.clicked.connect(self.abort)
        self._inputDock.widget.dxLE.editingFinished.connect(self._updateShiftValue)
        self._inputDock.widget.dyLE.editingFinished.connect(self._updateShiftValue)
        self._inputDock.widget._findShiftB.clicked.connect(self._findShift)
        self._chooseDimensionDock.widget.filterChanged.connect(self._filterStack)
        self._chooseDimensionDock.widget.stateDisabled.connect(self._wholeStack)

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """
        Dataset setter. Saves the dataset and updates the stack with the dataset
        data

        :param Dataset dataset: dataset
        """
        self.parent = parent
        self.dataset = dataset
        self._update_dataset = dataset
        self.indices = indices
        self.bg_indices = bg_indices
        self.bg_dataset = bg_dataset
        self._inputDock.widget.correctionB.setEnabled(True)
        if self.dataset.title != "":
            self._sv.setTitleCallback(lambda idx: self.dataset.title)
        if len(self.dataset.data.shape) > 3:
            self._chooseDimensionDock.show()
            self._chooseDimensionDock.widget.setDimensions(self._update_dataset.dims)
        if not self._chooseDimensionDock.widget._checkbox.isChecked():
            self._wholeStack()

    def getDataset(self, update=None):
        return self._update_dataset, self.indices, self.bg_indices, self.bg_dataset

    def _updateDataset(self, widget, dataset):
        self.parent._updateDataset(widget, dataset)
        self.dataset = dataset

    def correct(self):
        """
        Function that starts the thread to compute the shift given
        at the input widget
        """
        dx = self._inputDock.widget.getDx()
        dy = self._inputDock.widget.getDy()
        self.shift = numpy.array([dy, dx])
        if self._filtered_shift is None or self._inputDock.widget.checkbox.isChecked():
            frames = numpy.arange(
                self._update_dataset.get_data(
                    indices=self.indices, dimension=self._dimension
                ).shape[0]
            )
            self.thread_correction = OperationThread(
                self, self._update_dataset.apply_shift
            )
            self.thread_correction.setArgs(
                numpy.outer(self.shift, frames), self._dimension, indices=self.indices
            )
        else:
            self.thread_correction = OperationThread(
                self, self._update_dataset.apply_shift_along_dimension
            )
            self.thread_correction.setArgs(
                self._filtered_shift, self._dimension[0], indices=self.indices
            )
        self.thread_correction.finished.connect(self._updateData)
        self._inputDock.widget.correctionB.setEnabled(False)
        self._inputDock.widget.abortB.show()
        self.thread_correction.start()
        self.computingSignal.emit(True)

    def abort(self):
        self._inputDock.widget.abortB.setEnabled(False)
        self._update_dataset.stop_operation(Operation.SHIFT)

    def resetStack(self):
        self._update_dataset = self.dataset
        self.setStack()

    def updateProgress(self, progress):
        self.sigProgressChanged.emit(progress)

    def _findShift(self):
        if self._filtered_shift is not None:
            self.thread_detection = OperationThread(
                self, self._update_dataset.find_shift_along_dimension
            )
            self.thread_detection.setArgs(self._dimension[0], indices=self.indices)
        else:
            self.thread_detection = OperationThread(
                self, self._update_dataset.find_shift
            )
            self.thread_detection.setArgs(self._dimension, indices=self.indices)
        self._inputDock.widget._findShiftB.setEnabled(False)
        self.thread_detection.finished.connect(self._updateShift)
        self.thread_detection.start()
        self.computingSignal.emit(True)

    def _updateShiftValue(self):
        if self._filtered_shift is not None:
            self._filtered_shift[self._dimension[1]] = [
                self._inputDock.widget.getDy(),
                self._inputDock.widget.getDx(),
            ]

    def _updateShift(self):
        self._inputDock.widget._findShiftB.setEnabled(True)
        self.thread_detection.finished.disconnect(self._updateShift)
        if self._filtered_shift is None:
            self.shift = numpy.round(self.thread_detection.data[:, 1], 5)
        else:
            self._filtered_shift = []
            for s in self.thread_detection.data:
                try:
                    self._filtered_shift.append(numpy.round(s[:, 1], 5))
                except (IndexError, TypeError):
                    self._filtered_shift.append([0, 0])
            self._filtered_shift = numpy.array(self._filtered_shift)
            self.shift = self._filtered_shift[self._dimension[1][0]]
        self.computingSignal.emit(False)

    def _updateData(self):
        """
        Updates the stack with the data computed in the thread
        """
        self.thread_correction.finished.disconnect(self._updateData)
        self._inputDock.widget.abortB.hide()
        self._inputDock.widget.abortB.setEnabled(True)
        self._inputDock.widget.correctionB.setEnabled(True)
        self.computingSignal.emit(False)
        if self.thread_correction.data:
            del self._update_dataset
            self._update_dataset = self.thread_correction.data
            self.thread_correction.data = None
            self.thread_correction.func = None
            assert self._update_dataset is not None
            self.setStack(self._update_dataset)
        else:
            print("\nCorrection aborted")

    def setStack(self, dataset=None):
        """
        Sets new data to the stack.
        Mantains the current frame showed in the view.

        :param Dataset dataset: if not None, data set to the stack will be from the given dataset.
        """
        if dataset is None:
            dataset = self.dataset
        nframe = self._sv.getFrameNumber()
        data = (
            dataset.get_data(self.indices, self._dimension)
            if dataset is not None
            else None
        )
        self._sv.setStack(data if len(data) else None)
        self._sv.setFrameNumber(nframe)

    def clearStack(self):
        self._sv.setStack(None)
        self._inputDock.widget.correctionB.setEnabled(False)

    def _filterStack(self, dim=0, val=0):
        self._dimension = [dim, val]

        data = self._update_dataset.get_data(self.indices, self._dimension)
        if self.dataset.dims.ndim == 2:
            stack_size = self.dataset.dims.get(dim[0]).size
            reset_shift = (
                self._filtered_shift is None
                or self._filtered_shift.shape[0] != stack_size
            )
            self._inputDock.widget.checkbox.show()
            self._filtered_shift = (
                numpy.zeros((stack_size, 2)) if reset_shift else self._filtered_shift
            )
            self.shift = self._filtered_shift[val[0]]

        self._sv.setStack(data if len(data) else None)

    def _wholeStack(self):
        self._dimension = None
        self._filtered_shift = None
        self.shift = numpy.array([0, 0])
        self._inputDock.widget.checkbox.hide()
        self.setStack(self._update_dataset)

    def getStack(self):
        """
        Stack getter

        :returns: StackViewMainWindow:
        """
        return self._sv

    def getStackViewColormap(self):
        """
        Returns the colormap from the stackView

        :rtype: silx.gui.colors.Colormap
        """
        return self._sv.getColormap()

    def setStackViewColormap(self, colormap):
        """
        Sets the stackView colormap

        :param colormap: Colormap to set
        :type colormap: silx.gui.colors.Colormap
        """
        self._sv.setColormap(colormap)

    @property
    def shift(self):
        return self._shift

    @shift.setter
    def shift(self, shift):
        self._shift = numpy.array(shift)
        self._inputDock.widget.setDx(shift[1])
        self._inputDock.widget.setDy(shift[0])


class _InputDock(qt.QDockWidget):
    def __init__(self, parent=None):
        qt.QDockWidget.__init__(self, parent)
        self.widget = _InputWidget()
        self.setWidget(self.widget)


class _InputWidget(qt.QWidget):
    """
    Widget used to obtain the double parameters for the shift correction.
    """

    def __init__(self, parent=None):
        super(_InputWidget, self).__init__(parent)

        self._findShiftB = qt.QPushButton("Find shift")
        labelx = qt.QLabel("Horizontal shift:")
        labely = qt.QLabel("Vertical shift:")
        self.dxLE = qt.QLineEdit("0.0")
        self.dyLE = qt.QLineEdit("0.0")
        self.correctionB = qt.QPushButton("Correct")
        self.abortB = qt.QPushButton("Abort")
        self.abortB.hide()
        self.checkbox = qt.QCheckBox("Apply only to selected value")
        self.checkbox.setChecked(False)
        self.checkbox.hide()

        self.dxLE.setValidator(qt.QDoubleValidator())
        self.dyLE.setValidator(qt.QDoubleValidator())

        layout = qt.QGridLayout()

        layout.addWidget(self._findShiftB, 0, 0, 1, 2)
        layout.addWidget(labelx, 1, 0)
        layout.addWidget(labely, 2, 0)
        layout.addWidget(self.dxLE, 1, 1)
        layout.addWidget(self.dyLE, 2, 1)
        layout.addWidget(self.correctionB, 4, 0, 1, 2)
        layout.addWidget(self.abortB, 4, 0, 1, 2)
        layout.addWidget(self.checkbox, 3, 1)

        self.setLayout(layout)

    def setDx(self, dx):
        """
        Set the shift in the x axis
        """
        self.dxLE.setText(str(dx))
        self.dxLE.setCursorPosition(0)

    def getDx(self):
        """
        Get the shift in the x axis

        :return float:
        """
        return float(self.dxLE.text())

    def setDy(self, dy):
        """
        Set the shift in the x axis
        """
        self.dyLE.setText(str(dy))
        self.dyLE.setCursorPosition(0)

    def getDy(self):
        """
        Get the shift in the y axis

        :return float:
        """
        return float(self.dyLE.text())



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "06/04/2020"

import logging

import numpy
from .base import Base

_logger = logging.getLogger(__file__)


class NMF(Base):
    """
    Non-Negative Matrix Factorization.

    Find two non-negative matrices whose product approximates the non-negative
    matrix data.
    """

    def _init_w(self):
        if self.W is None:
            Base._init_w(self)
        self.W = numpy.abs(self.W)

    def _init_h(self):
        if self.H is None:
            Base._init_h(self)
        self.H = numpy.abs(self.H)

    def _update_h(self):
        _logger.info("Updating H")
        V, W, H = self.data, self.W, self.H
        if self.indices is None:
            for column in range(0, self.H.shape[1], self._hstep):
                mult = numpy.matmul(W.T, V[:, column : column + self._hstep]) / (
                    numpy.matmul(
                        numpy.matmul(W.T, W), H[:, column : column + self._hstep]
                    )
                    + 10**-9
                )
                self.H[:, column : column + self._hstep] *= mult
        else:
            for column in range(0, self.H.shape[1], self._hstep):
                mult = numpy.matmul(
                    W.T, V[self.indices, column : column + self._hstep]
                ) / (
                    numpy.matmul(
                        numpy.matmul(W.T, W), H[:, column : column + self._hstep]
                    )
                    + 10**-9
                )
                self.H[:, column : column + self._hstep] *= mult

    def _update_w(self):
        _logger.info("Updating W")
        V, W, H = self.data, self.W, self.H
        if self.indices is None:
            for row in range(0, len(self.W), self._vstep):
                mult = numpy.matmul(V[row : row + self._vstep], H.T) / (
                    numpy.matmul(numpy.matmul(W[row : row + self._vstep], H), H.T)
                    + 10**-9
                )
                self.W[row : row + self._vstep] *= mult
        else:
            for row in range(0, len(self.W), self._vstep):
                indx = self.indices[row : row + self._vstep]
                mult = numpy.matmul(V[indx], H.T) / (
                    numpy.matmul(numpy.matmul(W[row : row + self._vstep], H), H.T)
                    + 10**-9
                )
                self.W[row : row + self._vstep] *= mult

    def fit_transform(
        self,
        H=None,
        W=None,
        max_iter=1000,
        compute_w=True,
        compute_h=True,
        vstep=100,
        hstep=1000,
        error_step=None,
    ):
        """
        Find the two non-negative matrices (H, W) using Lee and Seung's multiplicative update rule (https://proceedings.neurips.cc/paper/2000/file/f9d1152547c0bde01830b7e8bd60024c-Paper.pdf).
        The images are loaded from disk in chunks.

        :param H: If not None, used as initial guess for the solution.
        :type H: array_like, shape (n_components, n_features), optional
        :param W: If not None, used as initial guess for the solution.
        :type W: array_like, shape (n_samples, n_components)
        :param max_iter: Maximum number of iterations before timing out,
            defaults to 200
        :type max_iter: int, optional
        :param compute_w: If False, W is not computed.
        :type compute_w: bool, optional
        :param compute_h: If False, H is not computes.
        :type compute_h: bool, optional
        :param vstep: vertical size of the chunks to take from data.
            When updating W, `vstep` images are retrieved from disk per iteration,
            defaults to 100.
        :type vstep: int, optional
        :param hstep: horizontal size of the chunks to take from fata.
            When updating H, `hstep` pixels are retrieved from disk per iteration,
            defaults to 1000.
        :type hstep: int, optional
        :param error_step: If None, error is not computed, else compute error for
            every `error_step` iterations.
        :type error_step: Union[None,int], optional
        """

        self.H = H
        self.W = W
        self._vstep = vstep
        self._hstep = hstep

        _logger.info("Starting NMF algorithm")

        Base.fit_transform(
            self,
            max_iter=max_iter,
            error_step=error_step,
            compute_w=compute_w,
            compute_h=compute_h,
            norm="squared_frobenius",
        )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "12/04/2022"

import numpy

from silx.gui import qt
from silx.gui import icons
from silx.gui.colors import Colormap
from silx.gui.plot import ScatterView, StackView, Plot1D
import silx.resources

import darfix
from darfix.io.utils import write_components
from .utils import ChooseDimensionDock


class DisplayComponentsWidget(qt.QMainWindow):
    """
    Widget to display rocking curves and scattering plot from a set of components
    """

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        # Widget with the type of bss and the number of components to compute
        widget = qt.QWidget(self)
        self._sv_components = StackView(parent=self)
        self._sv_components.setColormap(
            Colormap(
                name=darfix.config.DEFAULT_COLORMAP_NAME,
                normalization=darfix.config.DEFAULT_COLORMAP_NORM,
            )
        )
        self._sv_components.setGraphTitle("Components")
        self._mixing_plots_widget = MixingPlotsWidget(self)
        self._mixing_plots_widget.activeCurveChanged.connect(self._setComponent)
        self._sv_components.sigFrameChanged.connect(self._mixing_plots_widget._setCurve)

        self.bottom_widget = qt.QWidget(self)
        layout = qt.QGridLayout()
        componentsLabel = qt.QLabel("Components")
        rockingCurvesLabel = qt.QLabel("Rocking curves")
        font = qt.QFont()
        font.setBold(True)
        componentsLabel.setFont(font)
        rockingCurvesLabel.setFont(font)
        rockingCurvesLabel.setFont(font)
        layout.addWidget(componentsLabel, 1, 0, 1, 2, qt.Qt.AlignCenter)
        layout.addWidget(rockingCurvesLabel, 1, 2, 1, 2, qt.Qt.AlignCenter)
        layout.addWidget(self._sv_components, 2, 0, 2, 2)
        layout.addWidget(self._mixing_plots_widget, 2, 3, 2, 2)
        self.saveB = qt.QPushButton("Save components")
        self.saveB.pressed.connect(self._saveComp)
        layout.addWidget(self.saveB, 4, 4, 1, -1)
        widget.setLayout(layout)
        self.setCentralWidget(widget)

    def hideButton(self):
        self._computeB.hide()

    def showButton(self):
        self._computeB.show()

    def setComponents(self, components, W, dimensions, values, title=""):
        """
        Components setter. Updates the plots with the components and their
        corresponding rocking curves.

        :param array_like components: stack of images with the components
        :param array_like W: array with the rocking curves intensity
        :param dict dimensions: dictionary with the values of the dimensions
        """
        self.components = numpy.round(components, 5)
        self.W = W
        self.values = values
        self.dimensions = dimensions
        self._sv_components.setStack(self.components)
        if title != "":
            self._sv_components.setTitleCallback(lambda idx: title)
        self._mixing_plots_widget.updatePlots(self.dimensions, self.W.T, values)

    def _saveComp(self):
        """
        Loads the file from a FileDialog.
        """
        fileDialog = qt.QFileDialog()

        fileDialog.setFileMode(fileDialog.AnyFile)
        fileDialog.setAcceptMode(fileDialog.AcceptSave)
        fileDialog.setOption(fileDialog.DontUseNativeDialog)
        fileDialog.setDefaultSuffix(".h5")
        if fileDialog.exec_():
            write_components(
                fileDialog.selectedFiles()[0],
                "entry",
                self.dimensions.to_dict(),
                self.W,
                self.components,
                self.values,
                1,
            )

    def _setComponent(self, index=None):
        if index is not None:
            status = self._sv_components.blockSignals(True)
            self._sv_components.setFrameNumber(index)
            self._sv_components.blockSignals(status)

    def setStackViewColormap(self, colormap):
        """
        Sets the stackView colormap

        :param colormap: Colormap to set
        :type colormap: silx.gui.colors.Colormap
        """
        self._sv_components.setColormap(colormap)

    def getStackViewColormap(self):
        """
        Returns the colormap from the stackView

        :rtype: silx.gui.colors.Colormap
        """
        return self._sv_components.getColormap()


class MixingPlotsWidget(qt.QMainWindow):
    """
    Widget to show rocking curves and reciprocal space maps based on the mixing matrix.
    """

    activeCurveChanged = qt.Signal(int)

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        # Widget with the type of bss and the number of components to compute
        widget = qt.QWidget(self)

        self._is2dimensional = False

        self._plot_rocking_curves = PlotRockingCurves()
        self._plot_rocking_curves.sigStateDisabled.connect(self._wholeStack)
        self._plot_rocking_curves.sigMotorChanged.connect(self._updateMotorAxis)
        self._plot_rocking_curves.sigActiveCurveChanged.connect(
            self._activeCurveChanged
        )

        self._rsm_scatter = ScatterView()
        self._rsm_scatter.hide()

        self._toolbar = qt.QToolBar(parent=self)
        self._toolbar.setIconSize(self._toolbar.iconSize() * 1.2)
        if "darfix" not in silx.resources._RESOURCE_DIRECTORIES:
            silx.resources.register_resource_directory("darfix", "darfix.resources")
        curves_icon = icons.getQIcon("darfix:gui/icons/curves")
        scatter_icon = icons.getQIcon("darfix:gui/icons/scatter")
        self.curves_action = qt.QAction(curves_icon, "Curves", self)
        self.curves_action.setCheckable(True)
        self.curves_action.setChecked(True)
        self.rsm_action = qt.QAction(scatter_icon, "Scatter", self)
        self.rsm_action.setCheckable(True)
        self.curves_action.toggled.connect(self._activateCurvesPlot)
        self.rsm_action.triggered.connect(self.curves_action.toggle)
        self._toolbar.addAction(self.curves_action)
        self._toolbar.addAction(self.rsm_action)
        self._toolbar.setOrientation(qt.Qt.Vertical)
        self._toolbar.hide()

        self.bottom_widget = qt.QWidget(self)
        layout = qt.QGridLayout()
        layout.addWidget(self._plot_rocking_curves, 1, 0, 2, 2)
        layout.addWidget(self._rsm_scatter, 1, 0, 2, 2)
        layout.addWidget(self._toolbar, 1, 3, 2, 1)
        widget.setLayout(layout)
        self.setCentralWidget(widget)

    def is2dimensional(self, is2dimensional):
        self._is2dimensional = is2dimensional

    @property
    def scatter(self):
        return self._rsm_scatter

    def updatePlots(self, dimensions, mixing_matrix, dimensions_values):

        self.dimensions = dimensions
        self.mixing_matrix = mixing_matrix
        self.is2dimensional(True if dimensions.ndim == 2 else False)
        self._activateCurvesPlot(True)
        self.clearPlots()
        self.values = numpy.array(list(dimensions_values.values()))

        for i, row in enumerate(self.mixing_matrix):
            self._plot_rocking_curves.getPlot().addCurve(
                numpy.arange(len(row)), row, legend=str(i)
            )

        if self._is2dimensional:
            colormap = Colormap(name="jet", normalization="linear")
            self._rsm_scatter.getPlotWidget().setGraphXLabel(
                self.dimensions.get(0).name
            )
            self._rsm_scatter.getPlotWidget().setGraphYLabel(
                self.dimensions.get(1).name
            )
            self._rsm_scatter.setData(
                self.values[0], self.values[1], self.mixing_matrix[0]
            )
            self._rsm_scatter.setColormap(colormap)
            self._rsm_scatter.resetZoom()
            self._toolbar.show()
        else:
            self._toolbar.hide()

        if self.dimensions.ndim > 1:
            # Motor dimension and index
            self.dimension = 0
            self.index = 0
            self._plot_rocking_curves._chooseDimensionDock.show()
            self._plot_rocking_curves._chooseDimensionDock.widget.setDimensions(
                self.dimensions
            )
        else:
            self._plot_rocking_curves._chooseDimensionDock.hide()

        self._plot_rocking_curves.getPlot().setActiveCurve("0")

    def _activateCurvesPlot(self, checked=False):
        if checked:
            self.rsm_action.setChecked(False)
            self._rsm_scatter.hide()
            self._plot_rocking_curves.show()
        else:
            self._plot_rocking_curves.hide()
            self._rsm_scatter.show()

    def _activeCurveChanged(self, prev_legend=None, legend=None):
        if legend:
            self.activeCurveChanged.emit(int(legend))
            if self._is2dimensional:
                self._rsm_scatter.setData(
                    self.values[0], self.values[1], self.mixing_matrix[int(legend)]
                )

    def _setCurve(self, index=-1):
        if index >= 0:
            status = self._plot_rocking_curves.blockSignals(True)
            self._plot_rocking_curves.getPlot().setActiveCurve(str(index))
            self._plot_rocking_curves.blockSignals(status)
            if self._is2dimensional:
                self._rsm_scatter.setData(
                    self.values[0], self.values[1], self.mixing_matrix[index]
                )

    def _updateFrameNumber(self, index):
        """Update the current plot.

        :param index: index of the frame to be displayed
        """
        self.index = index
        self.clearPlots()
        dim_values = numpy.take(
            self.values.reshape(self.dimensions.shape), self.index, self.dimension
        )
        for i, row in enumerate(self.mixing_matrix):
            W = numpy.take(
                row.reshape(self.dimensions.shape), self.index, self.dimension
            )
            self._plot_rocking_curves.addCurve(dim_values, W, legend=str(i))

    def _wholeStack(self):
        # Whole rocking curves are showed
        self._plot_rocking_curves.clear()
        for i, row in enumerate(self.mixing_matrix):
            self._plot_rocking_curves.addCurve(
                numpy.arange(len(row)), row, legend=str(i)
            )
        self._plot_rocking_curves.getPlot().setGraphXLabel("Image id")

    def _updateMotorAxis(self, dim=0, val=0):
        """
        Updates the motor to show the rocking curve from.

        :param int axis: 0 if no motor is chosen, else axis of the motor.
        """
        dimension = [dim, val]
        # Make sure dimension and value are lists
        if type(dimension[0]) is int:
            dimension[0] = [dimension[0]]
            dimension[1] = [dimension[1]]
        indx = numpy.arange(len(self.mixing_matrix[0])).reshape(self.dimensions.shape)
        # For every axis, get corresponding elements
        for i, j in sorted(zip(dimension[0], dimension[1])):
            # Flip axis to be consistent with the data shape
            axis = self.dimensions.ndim - i - 1
            # dim_values = numpy.array(list(self.dimensions.values())[dim])
            indx = indx.take(indices=j, axis=axis)
        self._plot_rocking_curves.clear()
        axis = 0
        for i in range(self.dimensions.ndim):
            if i not in dimension[0]:
                axis = i
        text = self.dimensions.get(axis).name
        self._plot_rocking_curves.getPlot().setGraphXLabel(text)
        if len(indx.shape) == 1:
            for i, row in enumerate(self.mixing_matrix):
                W = row[indx]
                self._plot_rocking_curves.addCurve(
                    self.values[axis][indx], W, legend=str(i)
                )

    def clearPlots(self):
        self._plot_rocking_curves.clear()


class PlotRockingCurves(qt.QMainWindow):
    """
    Widget to plot the rocking curves of the components. It can be filtered
    to show only the rocking curves of a certain moving dimension.
    """

    sigStateDisabled = qt.Signal()
    sigMotorChanged = qt.Signal(list, list)
    sigActiveCurveChanged = qt.Signal(object, object)

    def __init__(self, parent=None, dimensions=None):
        qt.QMainWindow.__init__(self, parent)

        self._plot = Plot1D()
        self._plot.setGraphTitle("Rocking curves")
        self._plot.setGraphXLabel("Image id")
        self._plot.sigActiveCurveChanged.connect(self.sigActiveCurveChanged)

        self._chooseDimensionDock = ChooseDimensionDock(self, vertical=False)
        self._chooseDimensionDock.hide()
        self._motors = qt.QWidget()
        self._motors.hide()
        layout = qt.QVBoxLayout()
        layout.addWidget(self._plot)

        centralWidget = qt.QWidget()
        centralWidget.setLayout(layout)
        self.setCentralWidget(centralWidget)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._chooseDimensionDock)

        self.addCurve = self._plot.addCurve
        self._chooseDimensionDock.widget.filterChanged.connect(self.sigMotorChanged)
        self._chooseDimensionDock.widget.stateDisabled.connect(self.sigStateDisabled)

    def clear(self):
        self._plot.clear()

    def getPlot(self):
        return self._plot



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "14/10/2021"

import logging
import os

from silx.gui import qt
from darfix.core.data_selection import load_process_data
from silx.gui.dialog.DataFileDialog import DataFileDialog

_logger = logging.getLogger(__file__)


class DatasetSelectionWidget(qt.QTabWidget):
    """
    Widget that creates a dataset from a list of files or from a single filename.
    It lets the user add the first filename of a directory of files, or to
    upload manually each of the files to be read.
    If both options are filled up, only the files in the list of filenames
    are read.
    """

    sigProgressChanged = qt.Signal(int)

    def __init__(self, parent=None):
        qt.QTabWidget.__init__(self, parent)

        self._ish5CB = qt.QCheckBox("Is hdf5?")

        # Raw data
        self._rawFilenameData = FilenameSelectionWidget(parent=self)
        self._rawFilesData = FilesSelectionWidget(parent=self)
        self._inDiskCB = qt.QCheckBox("Use data from disk", self)
        titleWidget = qt.QWidget(self)
        titleLayout = qt.QHBoxLayout()
        titleLabel = qt.QLabel("Workflow title:")
        self._titleLE = qt.QLineEdit("")
        titleLayout.addWidget(titleLabel)
        titleLayout.addWidget(self._titleLE)
        titleWidget.setLayout(titleLayout)
        rawData = qt.QWidget(self)
        rawData.setLayout(qt.QVBoxLayout())
        rawData.layout().addWidget(titleWidget)
        rawData.layout().addWidget(self._ish5CB)
        rawData.layout().addWidget(self._rawFilenameData)
        rawData.layout().addWidget(self._rawFilesData)
        rawData.layout().addWidget(self._inDiskCB)
        rawData.layout().addWidget(titleWidget)
        self.addTab(rawData, "raw data")

        self._isH5 = False
        self._inDisk = False

        # Dark data
        self._darkFilenameData = FilenameSelectionWidget(parent=self)
        self.addTab(self._darkFilenameData, "dark data")

        # Treated data
        self._treatedDirData = DirSelectionWidget(parent=self)
        self.addTab(self._treatedDirData, "treated data")

        self._dataset = None
        self.bg_dataset = None
        self.indices = None
        self.bg_indices = None

        self.getRawFilenames = self._rawFilesData.getFiles
        self.getRawFilename = self._rawFilenameData.getFilename
        self.getDarkFilename = self._darkFilenameData.getFilename
        self.setRawFilenames = self._rawFilesData.setFiles
        self.setRawFilename = self._rawFilenameData.setFilename
        self.setDarkFilename = self._darkFilenameData.setFilename
        self.getTreatedDir = self._treatedDirData.getDir
        self.setTreatedDir = self._treatedDirData.setDir

        self._ish5CB.stateChanged.connect(self.__isH5)
        self._inDiskCB.stateChanged.connect(self.__inDisk)

    def loadDataset(self) -> bool:
        """
        Loads the dataset from the filenames.
        """
        dark_filename = self._darkFilenameData.getFilename()
        root_dir = self._treatedDirData.getDir()
        filenames = self._rawFilesData.getFiles()
        if not filenames:
            filenames = self._rawFilenameData.getFilename()
        (
            self._dataset,
            self.indices,
            self.bg_indices,
            self.bg_dataset,
        ) = load_process_data(
            filenames,
            root_dir=root_dir,
            in_memory=not self._inDisk,
            dark_filename=dark_filename,
            isH5=self._isH5,
            title=self._titleLE.text(),
        )
        return self.dataset is not None and self.dataset.nframes != 0

    def _updateDataset(self, dataset):
        self._dataset = dataset

    @property
    def dataset(self):
        return self._dataset

    def getDataset(self):
        return self._dataset, self.indices, self.bg_indices, self.bg_dataset

    def updateProgress(self, progress):
        self.sigProgressChanged.emit(progress)

    def __isH5(self, isH5):
        self._isH5 = isH5
        self._rawFilenameData.isHDF5(isH5)
        if isH5:
            self._rawFilesData.hide()
        else:
            self._rawFilesData.show()

    def __inDisk(self, inDisk):
        self._inDisk = bool(inDisk)


class FilesSelectionWidget(qt.QWidget):
    """
    Widget used to get one or more files from the computer and add them to a list.
    """

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        self._files = []

        self.setLayout(qt.QVBoxLayout())
        self._table = self._init_table()
        self._addButton = qt.QPushButton("Add")
        self._rmButton = qt.QPushButton("Remove")
        self.layout().addWidget(self._table)
        self.layout().addWidget(self._addButton)
        self.layout().addWidget(self._rmButton)
        self._addButton.clicked.connect(self._addFiles)
        self._rmButton.clicked.connect(self._removeFiles)

    def _init_table(self):

        table = qt.QTableWidget(0, 1, parent=self)
        table.horizontalHeader().hide()
        # Resize horizontal header to fill all the column size
        if hasattr(table.horizontalHeader(), "setSectionResizeMode"):  # Qt5
            table.horizontalHeader().setSectionResizeMode(0, qt.QHeaderView.Stretch)
        else:  # Qt4
            table.horizontalHeader().setResizeMode(0, qt.QHeaderView.Stretch)

        return table

    def _addFiles(self):
        """
        Opens the file dialog and let's the user choose one or more files.
        """
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(qt.QFileDialog.ExistingFiles)

        if not dialog.exec_():
            dialog.close()
            return

        for file in dialog.selectedFiles():
            self.addFile(file)

    def _removeFiles(self):
        """
        Removes the selected items from the table.
        """
        selectedItems = self._table.selectedItems()
        if selectedItems is not None:
            for item in selectedItems:
                self._files.remove(item.text())
                self._table.removeRow(item.row())

    def addFile(self, file):
        """
        Adds a file to the table.

        :param str file: filepath to add to the table.
        """
        assert os.path.isfile(file)
        item = qt.QTableWidgetItem()
        item.setText(file)
        row = self._table.rowCount()
        self._table.setRowCount(row + 1)
        self._table.setItem(row, 0, item)
        self._files.append(file)

    def getFiles(self):
        return self._files

    def setFiles(self, files):
        """
        Adds a list of files to the table.

        :param array_like files: List to add
        """
        for file in files:
            self.addFile(file)

    def getDir(self):
        if len(self._files):
            return os.path.dirname(self._files[0])
        return None


class DirSelectionWidget(qt.QWidget):
    """
    Widget used to obtain a filename (manually or from a file)
    """

    dirChanged = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        self._dir = qt.QLineEdit("", parent=self)
        self._dir.editingFinished.connect(self.dirChanged)
        self._addButton = qt.QPushButton("Upload directory", parent=self)
        # self._okButton =  qt.QPushButton("Ok", parent=self)
        self._addButton.pressed.connect(self._uploadDir)
        # self._okButton.pressed.connect(self.close)
        self.setLayout(qt.QHBoxLayout())

        self.layout().addWidget(self._dir)
        self.layout().addWidget(self._addButton)
        # self.layout().addWidget(self._okButton)

    def _uploadDir(self):
        """
        Loads the file from a FileDialog.
        """
        fileDialog = qt.QFileDialog()
        fileDialog.setOption(qt.QFileDialog.ShowDirsOnly)
        fileDialog.setFileMode(qt.QFileDialog.Directory)
        if fileDialog.exec_():
            self._dir.setText(fileDialog.directory().absolutePath())
            self.dirChanged.emit()
        else:
            _logger.warning("Could not open directory")

    def getDir(self):
        return str(self._dir.text())

    def setDir(self, _dir):
        self._dir.setText(str(_dir))


class FilenameSelectionWidget(qt.QWidget):
    """
    Widget used to obtain a filename (manually or from a file)
    """

    filenameChanged = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self._isH5 = False
        self._filename = None
        self._filenameLE = qt.QLineEdit("", parent=self)
        self._addButton = qt.QPushButton("Upload file", parent=self)
        # self._okButton =  qt.QPushButton("Ok", parent=self)
        self._addButton.pressed.connect(self._uploadFilename)
        # self._okButton.pressed.connect(self.close)
        self.setLayout(qt.QHBoxLayout())

        self.layout().addWidget(self._filenameLE)
        self.layout().addWidget(self._addButton)
        # self.layout().addWidget(self._okButton)

    def isHDF5(self, isH5):
        self._isH5 = isH5

    def _uploadFilename(self):
        """
        Loads the file from a FileDialog.
        """
        if self._isH5:
            fileDialog = DataFileDialog()
            if fileDialog.exec_():
                self._filename = fileDialog.selectedDataUrl().path()
                self._filenameLE.setText(self._filename)
                self.filenameChanged.emit()
            else:
                _logger.warning("Could not open file")
        else:
            fileDialog = qt.QFileDialog()
            fileDialog.setFileMode(qt.QFileDialog.ExistingFile)
            if fileDialog.exec_():
                self._filenameLE.setText(fileDialog.selectedFiles()[0])
                self._filename = fileDialog.selectedFiles()[0]
                self.filenameChanged.emit()
            else:
                _logger.warning("Could not open file")

    def getFilename(self):
        return str(self._filenameLE.text())

    def setFilename(self, filename):
        self._filename = filename
        self._filenameLE.setText(str(filename))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "11/12/2020"

import numpy

from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Input, Output
from silx.gui.colors import Colormap
from darfix.gui.noiseRemovalWidget import NoiseRemovalDialog
from darfix.core.process import NoiseRemoval


class NoiseRemovalWidgetOW(OWWidget):
    """
    Widget that computes the background substraction from a dataset
    """

    name = "noise removal"
    icon = "icons/noise_removal.png"
    want_main_area = False
    ewokstaskclass = NoiseRemoval

    # Inputs
    class Inputs:
        dataset = Input("dataset", tuple)
        colormap = Input("colormap", Colormap)

    # Outputs
    class Outputs:
        dataset = Output("dataset", tuple)
        colormap = Output("colormap", Colormap)

    # Settings
    method = Setting(str())
    background_type = Setting(str())
    kernel_size = Setting(str())
    bottom_threshold = Setting(str())
    chunks = Setting(list())
    step = Setting(str())
    mask = Setting(list())

    def __init__(self):
        super().__init__()

        self._widget = NoiseRemovalDialog(parent=self)
        self.controlArea.layout().addWidget(self._widget)
        self._widget.okSignal.connect(self._sendSignal)

    @Inputs.dataset
    def setDataset(self, _input):
        if _input is not None:
            dataset, update = _input
            self._widget.setDataset(*dataset)
            if self.method:
                self._widget.mainWindow.method = self.method
            if self.background_type:
                self._widget.mainWindow.background = self.background_type
            if self.kernel_size:
                self._widget.mainWindow.size = self.kernel_size
            if self.bottom_threshold:
                self._widget.mainWindow.bottom_threshold = self.bottom_threshold
            if self.step:
                self._widget.mainWindow.step = self.step
            if self.chunks:
                self._widget.mainWindow.chunks = self.chunks
            if self.mask:
                self._widget.mainWindow.mask = self.mask

            if update is None:
                self.show()
            elif update != self:
                self.Outputs.dataset.send(
                    ((self,) + self._widget.mainWindow.getDataset(), update)
                )

    @Inputs.colormap
    def setColormap(self, colormap):
        self._widget.mainWindow.setStackViewColormap(colormap)

    def _updateDataset(self, widget, dataset):
        self._widget.mainWindow._updateDataset(widget, dataset)

    def _sendSignal(self):
        """
        Function to emit the new dataset.
        """
        self.information()
        self.method = self._widget.mainWindow.method
        self.background_type = self._widget.mainWindow.background
        self.kernel_size = self._widget.mainWindow.size
        self.step = self._widget.mainWindow.step
        self.chunks = self._widget.mainWindow.chunks
        self.bottom_threshold = self._widget.mainWindow.bottom_threshold
        if isinstance(self._widget.mainWindow.mask, numpy.ndarray):
            self.mask = self._widget.mainWindow.mask.tolist()
        else:
            self.mask = self._widget.mainWindow.mask
        self.Outputs.dataset.send(
            ((self,) + self._widget.mainWindow.getDataset(), None)
        )
        self.Outputs.colormap.send(self._widget.mainWindow.getStackViewColormap())
        self.close()
